% Fit a jump binary classifier to training data and cross-validate by
% inference results.
%
% (C) 2017 A. Bemporad, August 11, 2017

rng(1) % for reproducibility
addpath('../functions');

% Generate test data
T=10000; % number of training data
K=3;     % number of modes
n=8;     % dimension of x

Tv=10000; % number of validation data

Vtol=1e-8;     % the training algorithm stops when the fit cost doesn't decrease more than Vtol
N=5;           % number of executions of the fitting algorithm = # different initial guesses
maxiter=5000;  % max total number of iterations
update_loss=0; % do not update loss on mode sequence after each iteration during regression.
loss_criterion = 2; % 1 = use total loss to choose among the optima achieved for different initial conditions 
%                     2 = use fitting error only

X=zeros(T+Tv,n);
Y=zeros(T+Tv,1);
THtrue=randn(n,K);  % true parameters of jump model
THtrue=THtrue(:,1:K);
THtrue(1,:)=THtrue(1,:)./abs(THtrue(1,:)); % makes |th(1)|=1 to make the solution unique

Strue=zeros(T+Tv+1,1);  % true mode sequence
s=1; % s0
Strue(1)=s;
Nchange=500; % Change mode every Nchange steps

YN=.1;             % output noise intensity
TAU=10.^(-2:.1:2); % hyper-parameter multiplying mode penalty L(S)

VALUE1=zeros(numel(YN),numel(TAU));
VALUE2=VALUE1;
VALUE3=VALUE1;
VALUE4=VALUE1;
TIME=VALUE1;
TIMEV=VALUE1;

for i1=1:numel(YN)
    yn=YN(i1);
    
    fprintf('\n\nyn=%5.4f: (be patient ...)\n',yn);
    
    j=0;
    for i=1:T+Tv
        % mode transition
        j=j+1;
        if round(j/Nchange)==j/Nchange
            s=s+1;
            if s>K
                s=1;
            end
        end
        Strue(i+1)=s;
        
        X(i,:)=10*randn(1,n);
        Y(i)=sign(THtrue(:,s)'*X(i,:)'+yn*randn);
    end
    
    % Define loss functions
    Linit=0*ones(K,1)/K; % all s0 have same probability
    Lmode=0*(1:K)'/100; % regularization on mode
    rho=1e-5;    % regularization on Theta
    loss_type=2; % loss for binary classification max(0,1-y*theta*x)
    
    % mode transition cost = Ltrans(st,st-1)
    for i2=1:numel(TAU)
        trans_prob=.1;
        tau=TAU(i2);
        Ltrans=tau*(ones(K,K)-eye(K)); % mode transition cost = Ltrans(st,st-1)
        
        t1=tic;
        S0=ceil(K*rand(T+1,1)); % initial guess
        [TH,S,iters,loss,Yhat]=jm_fitN(X(1:T,:),Y(1:T),N,S0(1:T+1),rho,Linit,Lmode,Ltrans,...
            maxiter,Vtol,loss_type,update_loss,loss_criterion);
        t1=toc(t1);
        TIME(i1,i2)=t1;
                
        if iters>=maxiter
            warning('Maximum number of iterations exceeded');
        end
        
        % Normalize Theta
        nTH=ones(n,1)*abs(TH(1,:));
        TH=TH./nTH;
        
        % Order Theta to get unique mode definition
        [TH2,ii]=sortrows(TH');TH2=TH2';
        [THt2,iit]=sortrows(THtrue');THt2=THt2';
        
        S1t=zeros(T,1);
        S1f=zeros(T,1);
        for i=1:T
            S1t(i)=find(iit==Strue(i));
            S1f(i)=find(ii==S(i));
        end
        
        label_mismatch_fit=sum(Y(1:T)~=Yhat)/T*100;
        mode_mismatch_fit=sum(S1t~=S1f)/T*100;
        VALUE3(i1,i2)=label_mismatch_fit;
        VALUE4(i1,i2)=mode_mismatch_fit;
        
        nmf=sum(S1t~=S1f);
        mismatchf=nmf/T*100;
        
        %%%%%%%%%%%%%%%%%%%%%
        % Validation
        
        prediction_type=1; % one-step ahead prediction
        loss_type=2;
        t1=tic;
        [YV,SV,Lt,SV1]=jm_recursive_inference_mex(X(T+1:T+Tv,:),Y(T+1:T+Tv,:),TH,Linit,Lmode,Ltrans,loss_type,prediction_type);
        t1=toc(t1);
        TIMEV(i1,i2)=t1;
        
        % adjust ordering of modes
        S2t=zeros(Tv,1);
        S2v=zeros(Tv,1);
        for i=T+1:T+Tv
            S2t(i-T)=find(iit==Strue(i+1));
            S2v(i-T)=find(ii==SV(i-T));
        end
        
        label_mismatch=sum(YV~=Y(T+1:T+Tv))/Tv*100;
        mode_mismatch=sum(S2t~=S2v)/Tv*100;
        
        fprintf('  tau=%6.6f; label=%5.2f%%[val],%5.2f%%[fit]; mode=%5.2f%%[val],%5.2f%%[fit], iters=%d\n',...
            tau,label_mismatch,label_mismatch_fit,mode_mismatch,mode_mismatch_fit,iters);
        
        VALUE1(i1,i2)=label_mismatch;
        VALUE2(i1,i2)=mode_mismatch;
        
    end
end

% plot results
figure(1)
clf
linewidth=3.0;
ftsize=24;

subplot(121)
semilogx(TAU,VALUE3(i1,:),TAU,VALUE1(i1,:),'LineWidth',linewidth);
ylabel('% label mismatch');
subplot(122)
semilogx(TAU,VALUE4(i1,:),TAU,VALUE2(i1,:),'LineWidth',linewidth);
xlabel('\tau')
ylabel('% mode mismatch')

for i=1:2
    subplot(1,2,i);
    xlabel('\tau')
    grid on
    h=legend({'training','inference'},'Location','NorthEast','Interpreter','LaTeX');
    set(h,'FontSize',ftsize);
    h=get(gcf,'CurrentAxes');
    set(h,'FontSize',ftsize);
end

% try
%     set(gcf,'Color','none','Position',[105 244 1276 729]);
%     % http://www.mathworks.com/matlabcentral/fileexchange/23629
%     export_fig figure2 -pdf
% catch
% end
% 
% set(gcf,'Color',[1 1 1]);
% movefile('figure2.pdf','"../../../figures/example_jm_classifier.pdf"');
