% Fit a jump linear model to training data and cross-validate by inference results.
%
% (C) 2017 A. Bemporad, July 22, 2017

rng(1) % for reproducibility
addpath('../functions');

% Generate test data
K=3;      % number of models
n=20;     % dimension of x
T=10000;  % number of training data
Tv=10000; % number of validation data

Vtol=1e-8;     % the training algorithm stops when the fit cost doesn't decrease more than Vtol
N=5;           % number of executions of the fitting algorithm = # different initial guesses
maxiter=5000;  % max total number of iterations
update_loss=0; % do not update loss on model sequence after each iteration during regression.
loss_criterion = 2; % 1 = use total loss to choose among the optima achieved for different initial conditions 
%                     2 = use fitting error only

X=zeros(T+Tv,n);
Y=zeros(T+Tv,1);
THtrue=randn(n,K); % true parameters of jump model

Strue=zeros(T+Tv+1,1); % true mode sequence
s=1; % s0
Strue(1)=s;
trans_prob=.05; % transition probability

YN=[0,.01,.05,.1,.2];  % output noise intensity
TAU=[0 10.^(-7:.2:1)]; % hyper-parameter multiplying mode penalty L(S)

VALUE1=zeros(numel(YN),numel(TAU));
VALUE2=VALUE1;
TIME=VALUE1;
TIMEV=VALUE1;

for i1=1:numel(YN)
    yn=YN(i1);
    
    fprintf('\n\nyn=%5.4f:\n',yn);
    
    for i=1:T+Tv
        % mode transition
        a=rand;
        if a<=trans_prob*(K-1)
            snew=s;
            while s==snew
                s=ceil(K*rand); % change mode
            end
        end
        Strue(i+1)=s;
        
        X(i,:)=randn(1,n);
        Y(i)=THtrue(:,s)'*X(i,:)'+yn*randn;
    end
    
    % Define loss functions
    Linit=0*ones(K,1)/K; % all s0 have same probability
    %Linit=(1:K)'/100; % all s0 have same probability
    Lmode=0*(1:K)'/100; % regularization on mode
    rho=1e-5;    % regularization on Theta
    loss_type=1; % quadratic loss

    % mode transition cost = Ltrans(st,st-1)
    for i2=1:numel(TAU)
        tau=TAU(i2);
        Ltrans=tau*(-log(trans_prob*(ones(K,K)-eye(K))+(1-(K-1)*trans_prob)*eye(K))); % mode transition cost = Ltrans(st,st-1)
        
        t1=tic;
        S0=ceil(K*rand(T+1,1)); % initial guess
        [TH,S,iters,V,Yhat]=jm_fitN(X(1:T,:),Y(1:T),N,S0(1:T+1),rho,Linit,Lmode,Ltrans,...
            maxiter,Vtol,loss_type,update_loss,loss_criterion);
        t1=toc(t1);
        TIME(i1,i2)=t1;
        
        if iters>=maxiter
            warning('Maximum number of iterations exceeded');
        end
        
        % Order Theta to get unique mode definition
        [TH2,ii]=sortrows(TH');TH2=TH2';
        [THt2,iit]=sortrows(THtrue');THt2=THt2';
        
        %%%%%%%%%%%%%%%%%%%%%
        % Validation

        prediction_type=1; % one-step ahead prediction
        loss_type=1;
        t1=tic;
        [YV,SV,Lt,SV1]=jm_recursive_inference_mex(X(T+1:T+Tv,:),Y(T+1:T+Tv,:),TH,Linit,Lmode,Ltrans,loss_type,prediction_type);
        t1=toc(t1);
        TIMEV(i1,i2)=t1;
        
        % adjust ordering of modes
        S2t=zeros(Tv,1);
        S2v=zeros(Tv,1);
        for i=T+1:T+Tv
            S2t(i-T)=find(iit==Strue(i+1));
            S2v(i-T)=find(ii==SV(i-T));
        end

        E=YV-Y(T+1:T+Tv);
        fit=norm(E)^2/Tv;
        
        nm=sum(S2t~=S2v);
        mismatch=nm/Tv*100;
        
        fprintf('  tau=%5.4f, fit=%6.2f, mismatch=%5.2f%%\n',tau,fit,mismatch);
        
        VALUE1(i1,i2)=fit;
        VALUE2(i1,i2)=mismatch;
        
    end
end

TAUBEST=2*YN.^2; % best theoretical tau given yn

% plot results
figure(1)
clf
ftsize=24;
leg={};
y1min=4;
linewidth=3.0;
c=zeros(numel(YN),3);
for i1=1:numel(YN)
    subplot(121)
    h=semilogx(TAU,VALUE1(i1,:),'LineWidth',linewidth);
    hold on
    c(i1,:)=get(h,'Color');
    ylabel('fit')
    h=get(gcf,'CurrentAxes');
    set(h,'FontSize',ftsize);
    hold on

    subplot(122)
    h=semilogx(TAU,VALUE2(i1,:),'LineWidth',linewidth);
    ylabel('% mode mismatch')
    hold on
    leg{i1}=sprintf('$\\sigma_y$ = %3.2f',YN(i1));
    h=get(gcf,'CurrentAxes');
    set(h,'FontSize',ftsize);
end
subplot(121)
h=legend(leg,'Location','NorthWest','Interpreter','LaTeX');
set(h,'FontSize',ftsize);
subplot(122)
h=legend(leg,'Location','NorthWest','Interpreter','LaTeX');
set(h,'FontSize',ftsize);


for i1=1:numel(YN)
    subplot(121)
    y1=min(VALUE1(i1,:));
    semilogx([TAUBEST(i1) TAUBEST(i1)],[y1min y1+.3],'Color',c(i1,:),'LineWidth',linewidth,...
        'LineStyle','--');
    subplot(122)
    y1=min(VALUE2(i1,:));
    semilogx([TAUBEST(i1) TAUBEST(i1)],[10 y1+1],'Color',c(i1,:),'LineWidth',linewidth,...
        'LineStyle','--');
end

for i=1:2
    subplot(1,2,i);
    xlabel('\tau')
    grid on
    ax=axis;
    axis([1e-6 10 ax(3:4)]);
end

% try
%     set(gcf,'Color','none','Position',[105 244 1276 729]);
%     % http://www.mathworks.com/matlabcentral/fileexchange/23629
%     export_fig figure1 -pdf
% catch
% end
% set(gcf,'Color',[1 1 1]);
% movefile('figure1.pdf','../../../figures/example_jm_linear.pdf');
