% Fitting a Markov Jump Linear Model to training data.
%
% (C) 2017 A. Bemporad, October 2, 2017

rng(1) % for reproducibility
addpath('../functions');

% Generate test data
K=4; % number of models
nx=8;
nu=2;
n=nx+nu;

% Generate random transition matrix
Pi=rand(K,K)+2*K*eye(K);
for i=1:K
    Pi(i,:)=Pi(i,:)/sum(Pi(i,:));
end
Atrue=zeros(nx,nx,K);
Btrue=zeros(nx,nu,K);
for i=1:K
    [Atrue(:,:,i),Btrue(:,:,i)]=ssdata(drss(nx,1,nu));
end

% Order (A,B) to get unique mode definition. Take first row of [A B] for sorting
auxt=zeros(nx+nu,K);
for i=1:K
    auxt(:,i)=[Atrue(1,:,i) Btrue(1,:,i)];
end
[~,iit]=sortrows(auxt');
Atrue=Atrue(:,:,iit);
Btrue=Btrue(:,:,iit);
Pi=Pi(iit,iit);

T=50000;  % number of training data
Tv=50000; % number of validation data

Vtol=1e-8;     % the training algorithm stops when the fit cost doesn't decrease more than Vtol
N=5;           % number of executions of the fitting algorithm = # different initial guesses
maxiter=500;   % max total number of iterations
update_loss=2; % update loss on model sequence after the training algorithm is done.
loss_criterion = 2; % 1 = use total loss to choose among the optima achieved for different initial conditions 
%                     2 = use fitting error only

X=zeros(T+Tv,n);
Y=zeros(T+Tv,nx);

Strue=zeros(T+Tv+1,1);  % true mode sequence
s=1; % s0
Strue(1)=s;

% exciting input
%U=randn(T+Tv,nu);
U=zeros(T+Tv,nu);
U(1,:)=sign(randn(1,nu));
for i=2:T+Tv
    for j=1:nu
        U(i,j)=U(i-1,j);
        if rand>.8
            U(i,j)=-U(i-1,j);
        end
    end
end

YN=[0,.01,.05];    % output noise intensity
TAU=10.^(-2:.2:2); % hyper-parameter multiplying mode penalty L(S)

VALUE1=zeros(numel(YN),numel(TAU));
VALUE2=VALUE1;
VALUE3=VALUE1;
VALUE4=VALUE1;
TIME=VALUE1;
TIMEV=VALUE1;
ITERS=VALUE1;

for i1=1:numel(YN)
    yn=YN(i1);
    x=randn(nx,1); % initial state
    
    % Generate data
    for i=1:T+Tv
        
        % mode transition
        Pis=Pi(s,:); % current probability distribution for mode s
        s=find(rand<cumsum(Pis));s=s(1);
        Strue(i+1)=s;
        
        u=U(i,:)';
        X(i,:)=[x' u'];
        x=Atrue(:,:,s)*x+Btrue(:,:,s)*u+yn*randn(nx,1);
    end
    xend=x;
    
    % verify correctedness of sequence S
    fprintf('\n\nyn=%5.4f - Empirical vs true transition matrix: norm(P-Pi)=%g:\n',...
        yn,norm(Pi-jm_trans_prob(Strue,K)));
    
    % define loss functions
    Linit=0*ones(K,1)/K; 
    Lmode=0*(1:K)'/100; % regularization on mode
    rho=1e-5;    % regularization on Theta
    loss_type=1; % quadratic loss
    
    % we have to fit x_i(t+1) to A(i,:)*x(t)+B_i(i,:)*u(t), i=1,...,nx
    for i2=1:numel(TAU)
        tau=TAU(i2);
        Ltrans=tau*ones(K,K); % mode transition cost = Ltrans(st,st-1)
        
        A=zeros(nx,nx,K);
        B=zeros(nx,nu,K);
        
        t1=tic;
        S0=ceil(K*rand(T+1,1)); % initial guess
        [TH,S,iters,V,Yhat,Ltrans2,Linit2]=jm_fitN(X(1:T,:),X(2:T+1,1:nx),N,S0(1:T+1),...
            rho,Linit,Lmode,Ltrans,maxiter,Vtol,loss_type,update_loss,loss_criterion);
        t1=toc(t1);
        TIME(i1,i2)=t1;
        ITERS(i1,i2)=iters;
        
        for i3=1:nx
            for j=1:K
                A(i3,:,j)=TH(1:nx,j,i3);
                B(i3,:,j)=TH(nx+1:nx+nu,j,i3);
            end
        end
        
        % order (A,B) to get unique mode definition, by
        % taking first row of [A B] for sorting
        aux=zeros(nx+nu,K);
        for i=1:K
            aux(:,i)=[A(1,:,i) B(1,:,i)];
        end
        [~,ii]=sortrows(aux');
        A=A(:,:,ii);
        B=B(:,:,ii);
        
        % adjust ordering of modes
        S1=zeros(T,1);
        for i=1:T
            S1(i)=find(ii==S(i));
        end
        S=S1;
        
        % compute difference of models (Frobenius norm)
        aux=0;
        for k1=1:K
            for k2=1:nx
                for k3=1:nx
                    aux=aux+(A(k2,k3,k1)-Atrue(k2,k3,k1))^2;
                end
                for k3=1:nu
                    aux=aux+(B(k2,k3,k1)-Btrue(k2,k3,k1))^2;
                end
            end
        end
        aux=sqrt(aux);
        [Pi2,p02]=jm_trans_prob(S,K); % a postieriori distributions
        VALUE3(i1,i2)=aux;
        VALUE4(i1,i2)=norm(Pi-Pi2);
        
        %%%%%%%%%%%%%%%%%%%%%
        % Validation

        prediction_type=1; % one-step ahead prediction
        loss_type=1;
        t1=tic;
        [YV,SV,Lt,SV1]=jm_recursive_inference_mex(X(T+1:T+Tv,:),[X(T+2:T+Tv,1:nx);xend'],TH,Linit2,Lmode,Ltrans2,loss_type,prediction_type);
        t1=toc(t1);
        TIMEV(i1,i2)=t1;
        
        SV1=zeros(Tv,1);
        for i=T+1:T+Tv
            SV1(i-T)=find(ii==SV(i-T));
        end
        SV=SV1;
        
        E=YV-[X(T+2:T+Tv,1:nx);xend'];
        fit=sum(sum(E.^2))/Tv;
        
        nm=sum(SV~=Strue(T+2:T+Tv+1));
        mismatch=nm/Tv*100;
        
        fprintf('  tau=%6.6f, fit=%6.2f, mismatch=%5.2f%%, iters = %d, diff models=%5.4g, diff Pi=%5.4f\n',...
            tau,fit,mismatch,iters,VALUE3(i1,i2),VALUE4(i1,i2));
        
        VALUE1(i1,i2)=fit;
        VALUE2(i1,i2)=mismatch;
    end
end

% plot results
figure(1)
clf
ftsize=24;
leg={};
y1min=4;
linewidth=3.0;
c=zeros(numel(YN),3);
for i1=1:numel(YN)
    subplot(121)
    h=semilogx(TAU,VALUE1(i1,:),'LineWidth',linewidth);
    hold on
    c(i1,:)=get(h,'Color');
    ylabel('fit')
    h=get(gcf,'CurrentAxes');
    set(h,'FontSize',ftsize);
    hold on
    
    subplot(122)
    h=semilogx(TAU,VALUE2(i1,:),'LineWidth',linewidth);
    ylabel('% mode mismatch')
    hold on
    leg{i1}=sprintf('$\\sigma_y$ = %3.2f',YN(i1));
    h=get(gcf,'CurrentAxes');
    set(h,'FontSize',ftsize);
end
subplot(121)
h=legend(leg,'Location','NorthWest','Interpreter','LaTeX');
set(h,'FontSize',ftsize);
subplot(122)
h=legend(leg,'Location','NorthWest','Interpreter','LaTeX');
set(h,'FontSize',ftsize);

for i=1:2
    subplot(1,2,i);
    xlabel('\tau')
    grid on
    %ax=axis;
    %axis([1e-6 10 ax(3:4)]);
end


% try
%     set(gcf,'Color','none','Position',[105 244 1276 729]);
%     % http://www.mathworks.com/matlabcentral/fileexchange/23629
%     export_fig figure3 -pdf
% catch
% end
% 
% set(gcf,'Color',[1 1 1]);
% movefile('figure3.pdf','../../../figures/example_MJLS.pdf');
