% Identification of a simple PWARX system over a Voronoi partition
%
% (C) D. Piga, October 17, 2017
%     Revised A. Bemporad, October 21, 2017

close all

rng(100)  % for reproducibility
addpath('../functions'); 

% Generate data
Tbig = 10000;
u = .1*randn(Tbig,1);
%u = randn(Tbig,1);
y = zeros(Tbig,1);
Sg = zeros(Tbig,1);

a1t = [.1 .2]; % True model parameters
a2t = [.5 .4];
b1t = [-0.4  0.1];
b2t = [0.3  0.4];
% a1t = [-0.5 0.3]; % True model parameters
% a2t = [0.3  0.2];
% b1t = [-0.4  0.1];
% b2t = [0.3  0.4];

for t = 3:Tbig
    if y(t-1)+0.5*y(t-2)-0.3*u(t-1)+2*u(t-2)+0.2>=0
        Sg(t)=1; %active mode
    else
        Sg(t)=2; %active mode
    end
    y(t) = a1t(1,Sg(t))*y(t-1)+a2t(1,Sg(t))*y(t-2)+b1t(1,Sg(t))*u(t-1)+b2t(1,Sg(t))*u(t-2);
end

THtrue=[a1t;a2t;b1t;b2t;0 0];

%%%%%%%%%%%%%%
% Parameters
loss_type=4;     % PWA identification with Voronoi diagram: quadratic loss
T_id = Tbig/2;   % length of training set
T_cal = Tbig/4;
TAU = [0 logspace(-6,0,19)];      % hyper-parameter multiplying initial and transition mode penalties
TAU_PWA = logspace(-2,1,20);  % Vector with hyper-parameter in the loss function 
%                  ||y-theta_{y,s}*[x;1]|| + tau_PWA*||x-theta_{x,s}||. 
%                  Cross validation is performed to choose the optimal tau_PWA
K = 2;           % Number of modes
rho=1e-5;        % regularization on Theta

na = 2;          % y(t) depends on y(t-1)...y(t-na)
nb = 2;          % y(t) depends on u(t-1)...u(t-nb)

Vtol=1e-8;       % the training algorithm stops when the fit cost doesn't decrease more than Vtol
N=5;             % number of executions of the fitting algorithm = # different initial guesses
maxiter=1000;    % max total number of iterations
update_loss=2;   % update loss on model sequence after the training algorithm is done.       
loss_criterion = 1; % 1 = use total loss to choose among the optima achieved for different initial conditions 
%                     2 = use fitting error only

% Decimate and extract training, calibration and validation data
u_id= u(1:T_id,:);
y_id= y(1:T_id,:);

u_cal= u(T_id+1:T_id+T_cal,:);
y_cal= y(T_id+1:T_id+T_cal,:);

u_val= u(T_id+T_cal+1:end,:);
y_val= y(T_id+T_cal+1:end,:);

Sg_val = Sg(T_id+T_cal+1:end,:);

T_val = size(u_val,1); % length of validation dataset

% Construct regressor matrix
nin = max(na,nb); % size of initial conditions (used for constructing the regressor matrix)
T = T_id-nin; % Size of regression matrix X
X = zeros(T,na+nb+1);
Y = zeros(T,1);
for t=nin+1:T_id
   X(t-nin,:) = [y_id(t-1:-1:t-na)' u_id(t-1:-1:t-nb)' 1];
   Y(t-nin,1) = y_id(t); 
end
 
% regressor matrix used after calibration
T_idtot = T_id+T_cal-nin; % Size of regression matrix X_idtot
u_idtot= u(1:T_id+T_cal,:);
y_idtot= y(1:T_id+T_cal,:);
X_idtot = zeros(T,na+nb+1);
Y_idtot = zeros(T,1);
for t=nin+1:T_id+T_cal
   X_idtot(t-nin,:) = [y_idtot(t-1:-1:t-na)' u_idtot(t-1:-1:t-nb)' 1];
   Y_idtot(t-nin,1) = y_idtot(t); 
end

% Initialize mode sequence and transition matrix 
S0=ceil(K*rand(T_idtot+1,1));  % initial guess on mode sequence
Pi=ones(K,K)/K; % initial guess on transition matrix
p0 = ones(K,1)/K;

%%%%%%%%%%%%%%
% Training phase
prediction_type = 2; % open-loop simulation
fprintf('Training started.........\n')
tstart = tic;
err_min = Inf;
    
ii=0;
for tau=TAU
    % define loss functions
    Ltrans= -tau*log(Pi);
    Linit = -tau*log(p0);
    Lmode = zeros(K,1);
    err_sim=zeros(numel(TAU_PWA),1);
    
    for indtau_PWA =1:numel(TAU_PWA)
        ii=ii+1;
        tau_PWA = TAU_PWA(indtau_PWA);
        [TH1,S,iters,V,Yhat,~,~,c]=jm_fitN(X(1:T,:),...
            Y,N,S0(1:T+1),rho,Linit,Lmode,Ltrans,maxiter,Vtol,loss_type,...
            update_loss,loss_criterion,tau_PWA);
        [yhat_sim, S_sim] = PWA_prediction_mex(u_cal,TH1,c,na,nb,...
            prediction_type,loss_type,y_cal); % Cross-validation
        err_sim(indtau_PWA) = norm(yhat_sim - y_cal); % Compute norm of the estimation error
        if err_sim(indtau_PWA)<err_min
            err_min = err_sim(indtau_PWA);
            tau_opt=tau;
            tau_PWA_opt = tau_PWA;
            BFR_min= 100*(1-err_min/norm(mean(y_cal) - y_cal));
        end
        fprintf('[%2d/%2d] tau=%10g, tau_PWA= %10g, BFR=%4.2f%%, best tau''s= [%10g,%10g]\n',...
            ii,numel(TAU_PWA)*numel(TAU),tau,tau_PWA,BFR_min,tau_opt,tau_PWA_opt);
    end
end
Ltrans= -tau_opt*log(Pi);
Linit = -tau_opt*log(p0);
Lmode = zeros(K,1);
[TH,S_opt,iters,V,Yhat,Ltrans2,Linit2,c]=jm_fitN(X_idtot(1:T_idtot,:),...
    Y_idtot,N,S0(1:T_idtot+1),rho,Linit,Lmode,Ltrans,maxiter,Vtol,...
    loss_type,update_loss,1,tau_PWA_opt);

elapsed_time = toc(tstart);
fprintf('Training finished............\n')
fprintf('Required time for training: %4.2f seconds\n',elapsed_time)

%%%%%%%%%%%%%%
% Validation

prediction_type = 1; % one-step ahead
[yhat_pred, S_pred] = PWA_prediction(u_val,TH,c,na,nb,prediction_type,loss_type,y_val);
err_pred = norm(yhat_pred - y_val); % Compute norm of the estimation error
BFR_pred = 100*(1-err_pred/norm(mean(y_val) - y_val));

prediction_type = 2; % open-loop simulation
[yhat_sim, S_sim] = PWA_prediction(u_val,TH,c,na,nb,prediction_type,loss_type,y_val);
err_sim = norm(yhat_sim - y_val); % Compute norm of the estimation error
BFR_sim = 100*(1-err_sim/norm(mean(y_val) - y_val));
fprintf('BFR in Validation: %4.2f %%\n',BFR_sim)

%%%%%%%%%%%%%%
% Plot results
NN=100;
subplot(211)
plot(yhat_sim(1:NN)-y_val(1:NN));
legend('difference')
grid
subplot(212)
plot(yhat_sim(1:NN));
hold on
plot(y_val(1:NN),'r')
legend('simulated output', 'actual output')
grid

% Order Theta to get unique mode definition
[TH2,ii]=sortrows(TH');TH2=TH2';
[THt2,iit]=sortrows(THtrue');THt2=THt2';

fprintf('True Theta:\n');
THt2

fprintf('Estimated Theta:\n');
TH2
