function [yhat, Shat] = PWA_prediction(u,TH,c,na,nb,prediction_type,loss_type,y)
%#codegen
% Simulate SISO PWA dynamical models with Voronoi partitioning.
%
% [yhat, Shat, norm_err] = PWA_prediction(u,TH,c,na,nb,prediction_type,y)
%
% Input arguments:
% ----------------
% u:  u(1),...,u(T)       [T-by-1 array]
% TH: model parameters    [n-by-K array]
% c:  centroids           [K-by-n array]
% na: y(t) will depend on y(t-1)...y(t-na)  [positive scalar]
% nb: y(t) will depend on u(t-1)...u(t-nb)  [positive scalar]
% prediction_type:        1: one-step ahead prediction
%                         2: open-loop simulation
% loss_type:              4: PWA on Voronoi partition
% y: measured output [T-by-1 array]. Used in one-step ahead prediction. Only initial
%    conditions are used in open-loop simulation.
%
% Output arguments:
% ----------------
% yhat: estimated output
% Shat: estimated mode sequence
%
% (C) D. Piga, October 15, 2017

T_val = size(u,1);
nin = max(na,nb);
K = size(TH,2);
yhat = y;

Shat =  ones(T_val,1);
dist = zeros(K,1);

switch prediction_type
    case 1 % one-step ahead prediction
        for ind = nin+1:T_val
            X_val = [y(ind-1:-1:ind-na)' u(ind-1:-1:ind-nb)' 1]; % Construct regressor
            
            if loss_type == 1 % Voronoi
                for inds = 1:K % Compute distance from centroid
                    dist(inds) = norm(X_val-c(inds,:));
                end
            elseif loss_type == 2
                for inds = 1:K % Compute distance from centroid
                    dist(inds) = -X_val*c(inds,:)'; %TH_polytope(:,inds);
                end
            end
            [~,Shat(ind)] = min(dist); %Active mode
            yhat(ind) = X_val*TH(:,Shat(ind));
        end
    case 2 % open-loop simulation
        for ind = nin+1:T_val
            X_val = [yhat(ind-1:-1:ind-na)' u(ind-1:-1:ind-nb)' 1]; % Construct regressor
            
            %The part below is the same of the part for case 1.
            % It is replicated to avoid to check the value of the  prediction_type at each time step
            if loss_type == 4 % Voronoi
                for inds = 1:K % Compute distance from centroid
                    dist(inds) = norm(X_val-c(inds,:));
                end
            end
            [~,Shat(ind)] = min(dist); %Active mode
            yhat(ind) = X_val*TH(:,Shat(ind));
        end
end
