function th=jm_binary_classification(X,Y,rho)
%#codegen
% Solve the binary classification problem
%
% min sum max(0,1-Y(i)*theta'*X(i,:)') + rho*theta'*theta
%
% with respect to theta using quadratic programming. By default the
% function uses GUROBI's QP solver, otherwise QUADPROG (slower) if
% GUROBI is not available.
%
% theta=jm_binary_classification(X,Y,rho)
%
% Input arguments:
% ----------------
% X: x(1),...,x(T)       [T-by-nx array]
% Y: y(1),...,y(T)       [T-by-1 array]
% rho: coefficient for quadratic regularization of model parameter vector theta.
%
% Output argument:
% ----------------
% theta: parameter vector [nx-by-1 array]
%
% (C) 2017 A. Bemporad, July 26, 2017

[ni,n]=size(X);
nvar=n+ni;

A=[speye(ni) (Y*ones(1,n)).*X]; % epsil(i) >= 1-Y(i)*theta'*X(i,:)'
rhs=ones(ni,1);
f=[ones(ni,1);zeros(n,1)];
%f=zeros(ni+n,1); % Use this for squared max

lb=[zeros(ni,1);-Inf(n,1)];
ub=Inf(nvar,1);
x0=zeros(nvar,1);

try
    % Use Gurobi's QP solver
    Q=blkdiag(sparse(ni,ni),.5*rho*speye(n));
    %Q=blkdiag(.5*eye(ni),.5*rho*speye(n));  % Use this for squared max
    model=struct('Q',Q,'obj',f,'alpha',0,'A',A,...
        'rhs',rhs,'sense',char('>'*ones(ni,1)),'lb',lb,...
        'ub',ub,'vtype',char('C'*ones(nvar,1)),'start',x0);
    
    params=struct('outputflag',0);
    result=gurobi(model,params);
    sol=result.x;
catch
    % Use QUADPROG QP solver (slower)
    Q=blkdiag(sparse(ni,ni),rho*speye(n));
    %Q=blkdiag(eye(ni),rho*speye(n));  % Use this for squared max
    opt=struct('Display','off');
    sol=quadprog(Q,f,-A,-rhs,[],[],lb,ub,x0,opt);
end

th=sol(ni+1:ni+n);

