function [S,V]=jm_dp(X,Y,TH,Linit,Lmode,Ltrans,inference_flag,loss_type,tau_PWA,c)
%#codegen
% Dynamic programming for reconstructing optimal mode sequence.
%
% [S,V]=jm_dp(X,Y,TH,Linit,Lmode,Ltrans,inference_flag,loss_type,c)
%
% Input arguments:
% ----------------
% X: x(1),...,x(T)       [T-by-nx array]
% Y: y(1),...,y(T)       [T-by-ny array]
% TH: model parameters   [nx-by-K-by-ny array] for loss_type=1,2
%                        or [nx-by-K array] for loss_type=3
% Linit: penalty on initial mode s(0) [K-by-1 array]
% Lmode: penalty on mode s(k), k>0 [K-by-1 array]
% Ltrans: penalty on transitions [K-by-K array]
% inference_flag: 0 = includes terminal cost ell(x(T),y(T),theta_s(T))
%                 1 = does not count the final cost ell(x(T),y(T),theta_s(T)). 
%                     This is used for inference, as y(T) is not given.
% loss_type:      1 = least squares ||y-theta*x||_2^2          
%                 2 = binary classification max(0,1-y*theta*x) 
%                     (or sum_j(max(0,1-y(j)*theta_j*x)) in case of dim(y)>1)
%                 3 = ||y-theta||_2^2
%                 4 = PWA regression over Voronoi partition. The cost is
%                     ||y-theta_{y,s}*x||_2^2+tau_PWA*||x-theta_{x,s}||_2^2 
% tau_PWA: [scalar] hyper-parameter (only required for PWA regression)
% c:       [K-by-n matrix] clusters' centroids, defining Voronoi partition (used required for PWA regression)
%    
% 
% Output arguments:
% ----------------
% S: optimal mode sequence s(0),...,s(T)
% V: optimal cost
%
% (C) 2017 A. Bemporad, July 20-26, 2017
%          Revised by A. Bemporad, October 2, 2017
%          Revised by D. Piga, October 15, 2017
%          Revised by A. Bemporad, October 21, 2017

[T,n]=size(X);

if T==0
    [V,j]=min(Linit);
    S=j(1);
    return
end

K=numel(Linit);
ny=size(Y,2);

S=zeros(T+1,1);
xi=zeros(n,1);
yi=zeros(ny,1);

% Dynamic programming step (TH is fixed)
J=zeros(K,T);
I=zeros(K,T); % I(k,t)=j means that when s(t-1)=k then the best next state is s(t)=j
loss=0;

for i=T:-1:0
    if i>0
        xi=X(i,:)';
        yi=Y(i,:)';
    end
    for k=1:K
        if i==0
            [V,j]=min(J(:,2)+Ltrans(k,:)'); % possible ties in argmin are taken care of by MATLAB
            I(k,i+1)=j(1);
            J(k,1)=Linit(k)+V;
        else
            switch loss_type
                case {1,4}
                    loss=(yi(1)-TH(:,k,1)'*xi)^2;
                    for t=2:ny
                        loss=loss+(yi(t)-TH(:,k,t)'*xi)^2;
                    end
                    if loss_type == 4
                        loss = loss+tau_PWA*(xi-c(k,:)')'*(xi-c(k,:)'); 
                    end     
                case 2
                    loss=max(1-yi(1)*TH(:,k,1)'*xi,0);
                    for t=2:ny
                        loss=loss+max(1-yi(t)*TH(:,k,t)'*xi,0);
                    end
                case 3
                    loss=sum((yi-TH(:,k)).^2);
            end
            if i==T
                J(k,i+1)=Lmode(k)+(1-inference_flag)*loss;
            else
                [V,j]=min(J(:,i+2)+Ltrans(k,:)');
                I(k,i+1)=j(1);
                J(k,i+1)=Lmode(k)+loss+V;
            end
        end
    end
end

% Compute optimal sequence
[V,j]=min(J(:,1)); % get optimal cost and s(0)
S(1)=j(1);
for i=1:T
    S(i+1)=I(S(i),i); % get full optimal sequence
end