function [TH,S,kk,V,c]=jm_fit(X,Y,S,rho,Linit,Lmode,Ltrans,kmax,Vtol,loss_type,tau_PWA)
%#codegen
% Fit a jump model to training data.
%
% [TH,S,kk,V,c]=jm_fit(X,Y,S,rho,Linit,Lmode,Ltrans,kmax,Vtol,loss_type)
%
% Input arguments:
% ----------------
% X: x(1),...,x(T)       [T-by-nx array]
% Y: y(1),...,y(T)       [T-by-ny array]
% S: s(0),...,s(T) initial guess for mode sequence [(T+1)-by-1 array]
% rho: coefficient for quadratic regularization of model parameters Theta.
% Linit: penalty on initial mode s(0) [K-by-1 array]
% Lmode: penalty on mode s(k), k>0 [K-by-1 array]
% Ltrans: penalty on transitions [K-by-K array]
% kmax: maximum number of algorithm iterations
% Vtol: tolerance to stop iterations, when total loss function V(k)-V(k-1)<=Vtol
% loss_type:      1 = least squares ||y-theta*x||_2^2          
%                 2 = binary classification max(0,1-y*theta*x) 
%                     (or sum_j(max(0,1-y(j)*theta_j*x)) in case of dim(y)>1)
%                 3 = ||y-theta||_2^2
%                 4 = PWA regression over Voronoi partition. The cost is
%                     ||y-theta_{y,s}*x||_2^2+tau_PWA*||x-theta_{x,s}||_2^2 
% tau_PWA: [scalar] hyper-parameter (only required for loss_type=4)
%
% Output arguments:
% ----------------
% TH: model parameters. This has dimensions (nx,K,ny) for loss_type=1,2
%     or dimensions (nx,K) for loss_type=3
% S: s(0),...,s(T) mode sequence [(T+1)-by-1 array]
% kk: number of iterations
% V: final total loss function
% c: cluster centroids (only for PWA regression) [K-by-n matrix]
%
% (C) 2017 A. Bemporad, July 20, 2017
%          Revised by A. Bemporad, October 2, 2017
%          Revised by D. Piga, October 15, 2017 (added PWA regression)
%          Revised by A. Bemporad, October 21, 2017

coder.extrinsic('jm_binary_classification');

n=size(X,2);
ny=size(Y,2);
K=numel(Linit);

kk=0;
TH=zeros(n,K,ny);

V=Inf;
Vold=Inf;
srho=sqrt(rho);
inference_flag=0;

c=[];
if loss_type==4
    c = zeros(K,n); % initialize centroids
end

while (Vold-V>Vtol || kk==0) && kk<kmax
    % while (~all(S1==S) || kk==0) && kk<kmax % alternative criterion
    kk=kk+1;
    
    % Regression step (S is fixed)
    for k=1:K
        ii=(S(2:end)==k);
        m=sum(ii);
        if m>=n % make it larger than 0, otherwise TH(:,k) is badly estimated
            switch loss_type
                case {1,4}
                    for t=1:ny
                        TH(:,k,t)=[X(ii,:);srho*eye(n)]\[Y(ii,t);zeros(n,1)];
                    end
                    if loss_type == 4 % cluster centroids 
                        c(k,:) = sum(X(ii,:))/m;
                    end
                case 2
                    for t=1:ny
                        TH(:,k,t)=jm_binary_classification(X(ii,:),Y(ii,t),rho); % assumes here p=1
                    end
                case 3
                    TH(:,k)=sum(Y(ii,:))'/m; % this is the average                    
            end
            
        end
    end
    % Dynamic programming step (TH is fixed, c is also fixed if loss_type=4)
    Vold=V;
    switch loss_type
        case {1,2,3}
            [S,V]=jm_dp(X,Y,TH,Linit,Lmode,Ltrans,inference_flag,loss_type,0,0);
        case 4
            [S,V]=jm_dp(X,Y,TH,Linit,Lmode,Ltrans,inference_flag,loss_type,tau_PWA,c);
     end
end