function [TH,S,iters,loss,Yhat,Ltrans,Linit,c]=jm_fitN(X,Y,N,S,rho,Linit,Lmode,Ltrans,maxiter,Vtol,loss_type,update_loss,loss_criterion,tau_PWA)
% Fit a jump model to training data, trying from N different initial
% guesses and picking up the best result.
%
% [TH,S,iters,loss,Ltrans,Linit]=jm_fitN(X,Y,S,rho,Linit,Lmode,Ltrans,maxiter,Vtol,loss_type,update_loss)
%
% Input arguments:
% ----------------
% X: x(1),...,x(T)       [T-by-nx array]
% Y: y(1),...,y(T)       [T-by-ny array]
% N: number of executions of regression algorithm
% S: s(0),...,s(T) first initial guess for mode sequence [(T+1)-by-1 array].
%    The remaining initial guesses for S are generated randomly.
% rho: coefficient for quadratic regularization of model parameters Theta.
% Linit: penalty on initial mode s(0) [K-by-1 array]
% Lmode: penalty on mode s(k), k>0 [K-by-1 array]
% Ltrans: penalty on transitions [K-by-K array]
% maxiter: maximum number of algorithm iterations
% Vtol: tolerance to stop iterations, when total loss function V(k)-V(k-1)<=Vtol
% loss_type:      1 = least squares ||y-theta*x||_2^2
%                 2 = binary classification max(0,1-y*theta*x)
%                     (or sum_j(max(0,1-y(j)*theta_j*x)) in case of dim(y)>1)
%                 3 = ||y-theta||_2^2
%                 4 = PWA regression over Voronoi partition. The cost is
%                     ||y-theta_{y,s}*x||_2^2+tau_PWA*||x-theta_{x,s}||_2^2 
% update_loss: 1 = Ltrans and Linit are updated from empirical probability 
%                  distributions of mode transitions and initial mode at
%                  each of the N iterations
%              2 = only updated at the end based on the best mode sequence S
%              0 = not updated at all.
% loss_criterion: 1 = use the cost J to choose the best solution obtained from different initial conditions
%                 2 = choose the solution obtained from different initial conditions providing the least fitting error [default]
% tau_PWA: [scalar] hyper-parameter (only required for loss_type=4)
%
% Output arguments:
% ----------------
% TH: model parameters. This has dimensions (nx,K,ny) for loss_type=1,2
%     or dimensions (nx,K) for loss_type=3
% S: s(0),...,s(T) mode sequence [(T+1)-by-1 array]
% iters: number of iterations
% loss: loss function
% Yhat: outputs estimated from X, S and TH
% Ltrans, Linit: modified mode losses
% c: cluster's centroids (only for PWA regression) [K-by-n matrix]
%
% (C) 2017 A. Bemporad, October 2, 2017
%     Revised by D. Piga, October 15, 2017 (to consider PWA models)
%     Revised by A. Bemporad, October 21, 2017

if nargin < 14 || isempty(tau_PWA)
    tau_PWA = 0;
end
if nargin < 13 || isempty(loss_criterion)
    loss_criterion = 2;
end
[T,n]=size(X);
ny=size(Y,2);
K=numel(Linit);
loss=Inf;
iters=0;
TH=zeros(n,K,ny);
c = zeros(K,n);

if update_loss>0
    tau=zeros(K,1);
    for i=1:K
        tau(i)=sum(Ltrans(i,:));
    end
    tau0=sum(Linit);
end

% Repeat the optimization from N different initial guesses
j=0;
while j<N && iters<maxiter
    j=j+1;
    if j==1
        S0=S;
    else
        S0=ceil(K*rand(T+1,1)); % Initial guess
    end
    
    [TH1,S1,iters1,V,c1]=jm_fit_mex(X(1:T,:),Y(1:T,:),S0(1:T+1),rho,Linit,Lmode,Ltrans,maxiter,Vtol,loss_type,tau_PWA);
 
    % Compute loss function
    switch loss_criterion
        case 1
            [~,Yhat1]=jm_loss_mex(X,Y,S1,TH1,loss_type,tau_PWA,c1);
            loss1 = V;
        case 2    
            [loss1,Yhat1]=jm_loss_mex(X,Y,S1,TH1,loss_type,tau_PWA,c1);
    end

    if loss1<loss
        TH=TH1;
        loss=loss1;
        S=S1;
        Yhat=Yhat1;
        if loss_type == 4
            c=c1;
        end
    end
    iters=iters+iters1;
    maxiter=maxiter-iters1;
    
    if update_loss==1 || (update_loss==2 && j==N)
        
        %prediction_type=1;
        %[~,S]=jm_recursive_inference(X,Y,TH,Linit,Lmode,Ltrans,loss_type,prediction_type);
        
        [P,p]=jm_trans_prob_mex(S,K); % A postieriori distributions

        [Ltrans,Linit]=jm_mode_loss_mex(P,p,tau,tau0);

    end
end
    