function [Yhat,Shat,Lt,Shat1]=jm_recursive_inference(X,Y,TH,Linit,Lmode,Ltrans,loss_type,prediction_type)
%#codegen
% Recursive output estimation using jump model.
%
% [Yhat,Shat,Lt]=jm_recursive_inference(X,Y,TH,Linit,Lmode,Ltrans,loss_type,prediction_type)
%
% Input arguments:
% ----------------
% X: x(1),...,x(T)       [T-by-nx array]
% Y: y(1),...,y(T)       [T-by-ny array]
% TH: model parameters   [nx-by-K-by-ny array] for loss_type=1,2
%                        or [nx-by-K array] for loss_type=3
% Linit: penalty on initial mode s(0) [K-by-1 array]
% Lmode: penalty on mode s(k), k>0 [K-by-1 array]
% Ltrans: penalty on transitions [K-by-K array]
% loss_type:       1 = least squares ||y-theta*x||_2^2          
%                  2 = binary classification max(0,1-y*theta*x) 
%                      (or sum_j(max(0,1-y(j)*theta_j*x)) in case of dim(y)>1)
%                  3 = ||y-theta||_2^2 [X is not used]
% prediction_type: 1 = one-step ahead prediction
%                  2 = open-loop prediction
%
% Output arguments:
% ----------------
% Yhat:  predicted outputs y(t|t-1)  [T-by-ny array]
% Shat:  predicted modes shat(t|t-1) [T-by-1 array]
% Lt:    mode penalty after T steps  [K-by-1 array]
% Shat1: predicted modes shat(t|t)   [T-by-1 array]
%
% (C) 2017 A. Bemporad, July 22, 2017
%          Revised by A. Bemporad, October 2, 2017

T=size(X,1);
K=size(TH,2);
ny=size(Y,2);

Yhat=zeros(T,ny);
Shat=zeros(T,1);
Shat1=zeros(T,1);
y=zeros(ny,1);
v=zeros(K,1);

Lt=Linit;
for k=1:K
    Lt(k)=Lt(k)+Lmode(k)+min(Lt+Ltrans(k,:)');
end

for t=1:T

    j=1;
    Ltmin=Lt(1);
    for k=2:K
        if Lt(k)<Ltmin
            j=k;
            Ltmin=Lt(k);
        end
    end
    Shat(t)=j; % we assume here that shat(t) = argmin_k Lt(k), assuming that a value yhat(t)
    %            exists such that ell(x(t),yhat(t),theta_s(t))=0.
    
    % Compute Yhat
    switch loss_type
        case 1
            Yhat(t,1)=X(t,:)*TH(:,j,1);
            for k=2:ny
                Yhat(t,k)=X(t,:)*TH(:,j,k);
            end
        case 2
            Yhat(t,1)=sign(X(t,:)*TH(:,j,1));
            for k=2:ny
                Yhat(t,k)=sign(X(t,:)*TH(:,j,k)); % we should assign 0 to +1 or -1 here
            end
        case 3
            Yhat(t,:)=TH(:,j)'; 
    end
    
    % Update Lt for step t+1
    switch prediction_type
        case 1
            y=Y(t,:)';
        case 2
            y=Yhat(t,:)';
    end

    % Compute loss ell(X(t,:),y(t),TH(j)) for all j=1,...,K
    ell=zeros(K,1);
    for j=1:K
        switch loss_type
            case 1
                ell(j)=(y(1)-X(t,:)*TH(:,j,1))^2;
                for h=2:ny
                    ell(j)=ell(j)+(y(h)-X(t,:)*TH(:,j,h))^2;
                end
            case 2
                ell(j)=max(1-y(1)*X(t,:)*TH(:,j,1),0);
                for h=2:ny
                    ell(j)=ell(j)+max(1-y(h)*X(t,:)*TH(:,j,h),0);
                end
            case 3
                ell(j)=sum((y-TH(:,j)).^2);
        end
    end
    
    [~,Shat1(t)]=min(ell+Lt); % Shat1(t)=shat(t|t)=argmin {ell(X(t,:),y(t),TH(j)) + Lt(j)}

    % Lt(k)=Lmode(k)+min(Lt+ell(X(t,:),y,TH)+Ltrans(k,:)')
    for k=1:K
        for j=1:K
            v(j)=Lt(j)+ell(j)+Ltrans(k,j);
        end
        Lt(k)=Lmode(k)+min(v);
    end
end





